
#  Copyright 2003-2006 Frederick Dean

# 
#  This loads a table definition from the field_def table
#  and composes a list.pm object to display results to the user.
#  This file is also responsible for forming the "where" clause
#  of queries.
#

use strict;

package FKong::table;

use FKong::list;

sub field_id_with_link { 
   return FKong::cgi::Link("/$_[2]/$_[1]$_[0].html") . "$_[1]$_[0]</a>" 
};

sub delete_link { return FKong::cgi::Link("delete.html",["id=$_[0]"]) ."delete</a>" }

our %right_align = (mult_choice => 0, integer => 1, hexint => 1, money => 1, oneline => 0, multiline => 0, 
                   bool => 0, float => 1, constant => 0,user => 0, unix_ts => 1, ip => 0);

sub hex_format { return sprintf("0x%02x",$_[0]); }
sub money_format { 
   my $money = $_[0];
   my $neg = ($money =~ s/^-//);  # remove leading dash
   my $short = 1 + $FKong::config{'MoneyDecimalPlaces'} - length $money;
   $money = ("0" x $short) . $money if $short > 0;  # append leading zeros if needed
   $money =~ s/(.{$FKong::config{'MoneyDecimalPlaces'}})$/.$1/;   # insert decimal
   return $neg ? "-$money" : $money; 
}
sub yes_no { return $_[0] ? "yes" : (length($_[0]) ? '<font color="#888888">no</font>' : '') }

sub user_link { return FKong::session::has_priv('view_users') ? FKong::cgi::Link("/user/~$_[0].html") . FKong::cgi::htmlQuote($_[0]) ."</a>" : 
                                                    FKong::cgi::htmlQuote($_[0]) };

# This converts a database integer or other type to a human presentable form
our %type_func = ( unix_ts => \&FKong::list::unixTimestamp, hexint => \&hex_format, money => \&money_format, bool => \&yes_no,
                   ip => \&FKong::list::ipAddress, user => \&user_link,
                   multiuser => sub { FKong::record::translate_to_username($_[0], $_[3]) },
                 );

our %custom_func = ( "fktable.url" => sub { return FKong::cgi::Link("../$_[0]/list.html") . FKong::cgi::htmlQuote($_[0]) ."</a>"; } ,
                     "fktable.tablesql" => sub { return FKong::cgi::Link("../$_[0]/show_create.html") . FKong::cgi::htmlQuote($_[0]) ."</a>"; } ,
                     "field_def.fieldName" => sub { return $_[0]; } ,
                     "session.cookie" => sub { return "<font color=#888888>hidden</a>" } );

sub fktable_actions { return FKong::cgi::Link("../$_[0]/field/list.html") . "design</a>"; }

sub define_list_object
{
   my($tbl,$fktable) = @_;
   my $tablesql = $$fktable{'tablesql'};
   my $pkey = $$fktable{'pkey'};
   my $tablename = $$fktable{'name'};
   my $dbh = FKong::db::SendSQL("SELECT fieldId, list_show, fieldName, short_help, type, sqlName FROM field_def\n".
                                "WHERE tablesql = ". FKong::db::SqlQuote($tablesql) ."\n".
                                "AND type != 'constant' ORDER BY sort, fieldName");
   #           nick private         title  align default   sql          display      descript                       
   $tbl->DefCol("T00",0,                  "ID",1,1,"$tablesql.$pkey",\&field_id_with_link, "Internal database ID of the $tablesql table");
   $tbl->DefCol("T01",0,                  "ID",1,0,"$tablesql.$pkey",\&delete_link, "Link to delete") if FKong::session::has_priv('delete_features');
   $tbl->DefCol("T02",0,             "Actions",1,1,"$tablesql.url", \&fktable_actions, "Do stuff to the table") if $tablesql eq "fktable";
   while(my($fieldId,$list_show,$fieldName,$short_help,$type,$sqlName) = $dbh->fetchrow_array()) {
      my $join;
      if($type eq 'user') {
         $join = "LEFT JOIN user AS $sqlName ON ($tablesql.$sqlName = $sqlName.userid)\n";
         $sqlName = "$sqlName.username";
      } else {
         $sqlName = "$tablesql.$sqlName";
      };
      my $func = $type_func{$type};
      $func = $custom_func{$sqlName} if $custom_func{$sqlName};
      $tbl->DefCol("t$fieldId",0,$fieldName,$right_align{$type},$list_show,$sqlName,$func, $short_help,$join);
      if($type eq 'multiline') {
         $tbl->DefCol("t${fieldId}L",0,$fieldName ." Length",1,0,"LENGTH($sqlName)",$func, $short_help,$join);
      }
   };
   $tbl->DefCol("T60",0,            "Comments",1,1,"COUNT(${tablesql}_comment.commentId)", undef, "A count of the number of comments",
                "LEFT JOIN ${tablesql}_comment ON (${tablesql}_comment.$pkey = $tablesql.$pkey)\n") if $$fktable{'comments'};
}

# Normally we save the column choices when composing a list, but can miss that opportunity
# when redirecting to the search form, so this form forces that to happen an should be callled
# when change_columns form parameter is defined.
sub force_save_column_choice
{
   my($fktable) = @_;
   my $tbl = FKong::list->new($$fktable{'tablesql'},$$fktable{'recprefix'},$$fktable{'url'}, {});
   define_list_object($tbl,$fktable);
   $tbl->get_column_sql([],"-T00");
}

sub show_table
{
   my($keep_list,$fktable,@wheres) = @_;
   my $tablesql = $$fktable{'tablesql'};
   my $recprefix = $$fktable{'recprefix'};
   # check that they have permission 
   FKong::session::must_have_priv("view_features");
   FKong::session::must_have_priv("view_fields") if $$fktable{'internal'};
   my $tbl = FKong::list->new($tablesql,$recprefix,$$fktable{'url'}, {});
   $tbl->keep_list() if $keep_list;  # set flag to store recId list
   define_list_object($tbl,$fktable);
   # column headings
   my $keepers = [ FKong::list::column_params($tbl) ];    # list reference
   my($colsql,$ordersql,$joins) = $tbl->get_column_sql($keepers,"-T00");
   # maximum number of records stuff
   my $limit = FKong::db::sql_limit();
   # submit the query
   foreach my $param (sort keys %FKong::cgi::form) {
      next unless $param =~ /^t(\d+)$/;  # param must be a feature list param
      next unless $FKong::cgi::form{$param};   # param must be true
      my $fieldId = $1;
   };
   my $where = "";
   $where = "WHERE ". join("\nAND ",@wheres) ."\n" if scalar @wheres;
   my $query = "SELECT $keep_list $colsql\nFROM $tablesql\n".
               $joins .
               $where .
               "GROUP BY $tablesql.$$fktable{'pkey'} ORDER BY $ordersql\n$limit";
   my $template = FKong::cgi::first_existing("template/$tablesql-list.html","template/features-template.html");
   $tbl->print_template_with_table($template,$query,$keepers);
}

our @search_params;  # These are the URL GET parameter names which matter to the search.  Used for form_query_link()

sub form_query_link
{
   my $query_url = "";
   foreach my $param (@search_params,"search_comments","comment_author") {
      my $value = $FKong::cgi::form{$param};
      next unless defined $value && length $value;  # skip this param if empty
      $query_url .= "&" if $query_url;
      $query_url .= join('&',map { "$param=". FKong::cgi::urlQuote($_) } split(/\0/,$FKong::cgi::form{$param}));
      #$query_url .= "$param=". FKong::cgi::urlQuote($value);  # or am I confident a character nul will not cause problems?
   };
   return $query_url;
}

#  First we look through all the enum values.
#  If an enum value is listed then that field must be one of the values listed.
#  If two enum fields happen to allow the same value, we are broken.  FIXME
#  Any words which are not used for enum, are then matched for one of the text areas.
sub basic_query
{
   my($fktable,$suffix) = @_;
   # process free-form 'text' search criteria
   my %enums;
   my @extra_words;  # terms that do not match a multi-choice
   push(@search_params,'text');
   foreach my $word (split(/\s+/,$FKong::cgi::form{'text'} || "")) {
      my $dbh = FKong::db::SendSQL("SELECT sqlName FROM choice LEFT JOIN field_def USING (fieldId)\n".
                            "WHERE choice = ". FKong::db::SqlQuote($word));
      if(my $sqlName = $dbh->fetchrow_array()) {
         $enums{$sqlName} .= "\0" if $enums{$sqlName};
         $enums{$sqlName} .= $word;
      } else {
         push(@extra_words,$word);
      };
   }
   # Create SQL
   my @wheres;  # to keep all the AND'ed clauses
   foreach my $sqlName (keys %enums) {  # for every enum field
      my @choices = split(/\0/,$enums{$sqlName});
      push(@wheres,"(". join(" OR ",map { "$sqlName = ". FKong::db::SqlQuote($_) } @choices) .")");
   };
   # Now create SQL for extra words
   if(scalar(@extra_words)) {
      my $dbh = FKong::db::SendSQL("SELECT sqlName FROM field_def WHERE (adv_search OR basic_search)\n".
                                   "AND (type = 'oneline' OR type = 'multiline')\n".
                                   "AND tablesql = ". FKong::db::SqlQuote($$fktable{'tablesql'}));
      my @sqlNames;
      while(my $sqlName = $dbh->fetchrow_array()) {
         push(@sqlNames,$sqlName);
      };
      foreach my $word (@extra_words) {  # for every extra word
         $word = FKong::db::SqlQuote(lc($word));
         push(@wheres,"(". join(" OR ",map { "INSTR(LOWER($_),$word)" } @sqlNames) .")");
      };
   };
   # Process field_def field specifications
   my %usercache;
   my $dbh = FKong::db::SendSQL("SELECT sqlName, type, fieldName, adv_search && ! basic_search FROM field_def\n".
                                "WHERE (basic_search OR adv_search) && tablesql = ". FKong::db::SqlQuote($$fktable{'tablesql'}));
   my $has_adv;
   while(my($sqlName,$type,$fieldName,$adv_only) = $dbh->fetchrow_array()) {
      if($type eq "oneline" || $type eq "multiline") {
         next unless $FKong::cgi::form{$sqlName};
         push(@wheres,"INSTR(LOWER($sqlName),". FKong::db::SqlQuote(lc($FKong::cgi::form{$sqlName})) .")");
         push(@search_params,$sqlName);
      } elsif($type eq "bool") {
         next unless defined $FKong::cgi::form{$sqlName} && length $FKong::cgi::form{$sqlName};
         push(@wheres,$FKong::cgi::form{$sqlName} ? "$sqlName" : "! $sqlName");
         push(@search_params,$sqlName);
      } elsif($type eq "mult_choice") {
         next unless $FKong::cgi::form{$sqlName};
         push(@wheres,"$sqlName IN (". (join(',',map { FKong::db::SqlQuote($_) } split(/\0/,$FKong::cgi::form{$sqlName}))) .")");
         push(@search_params,$sqlName);
      } elsif($type eq "integer" || $type eq 'hexint' || $type eq 'ip' || $type eq 'money' || $type eq "float" || $type eq 'unix_ts'){# if min-max
         my $func = $FKong::submit::to_db_type_func{$type};
         my $found;
         if(defined $FKong::cgi::form{"$sqlName-min"} && length FKong::cgi::trim($FKong::cgi::form{"$sqlName-min"})) {
            $found = 1;
            my $min = $FKong::cgi::form{"$sqlName-min"};
            my $min = defined $func ? &$func($min) : $min;
            if(defined $min) {
               push(@wheres,"$sqlName >= ". FKong::db::SqlQuote($min));
               push(@search_params,"$sqlName-min");
            } else {
               $FKong::cgi::keyword{'warning'} .= "<div class=warn>". FKong::cgi::htmlQuote($fieldName) ." minimum not understood.</div>\n";
            }
         }
         if(defined $FKong::cgi::form{"$sqlName-max"} && length FKong::cgi::trim($FKong::cgi::form{"$sqlName-max"})) {
            $found = 1;
            my $max = $FKong::cgi::form{"$sqlName-max"};
            my $max = defined $func ? &$func($max) : $max;
            if(defined $max) {
               push(@wheres,"$sqlName <= ". FKong::db::SqlQuote($max));
               push(@search_params,"$sqlName-max");
            } else {
               $FKong::cgi::keyword{'warning'} .= "<div class=warn>". FKong::cgi::htmlQuote($fieldName) ." maximum not understood.</div>\n";
            }
         }
         next unless $found;
      } elsif($type eq 'user') {
         my $slang = $FKong::cgi::form{$sqlName};
         next unless $slang;
         my($userid, $problem) = FKong::submit::identify_user($slang,$fieldName,\%usercache);
         if($problem) {
            $FKong::cgi::keyword{'warning'} .= "<div class=warn>$problem</div>\n";
            push(@wheres, "0");  #  Ouch!
         } else {
            push(@wheres, $$fktable{'tablesql'} .".$sqlName = $userid");
         }
         push(@search_params,$sqlName);
      } else { # else do nothing for this type
         next;
      }
      # okay if reaching here the user must have entered something for this field.
      $has_adv ||= $adv_only;
   };
   $FKong::cgi::keyword{'oldQuery'} = form_query_link();
   $FKong::cgi::keyword{'oldQuery'} .= '&adv=1' if $has_adv;  # append suffix to indicate advanced
   FKong::session::set_state("$$fktable{'tablesql'}-$suffix", $FKong::cgi::keyword{'oldQuery'});
   return \@wheres;
}

sub delete_record
{
   my($fktable) = @_;
   my $tablesql = $$fktable{'tablesql'};
   my $pkey = $$fktable{'pkey'};
   FKong::session::must_have_priv('delete_features');
   FKong::session::must_have_priv("edit_fields") if $$fktable{'internal'};
   FKong::Fatal("Internal Error: delete has no id") if ! $FKong::cgi::form{'id'};
   FKong::Fatal("Internal Error: delete has malformed id") if $FKong::cgi::form{'id'} !~ /^\d+$/;
   FKong::Fatal("Sorry. You cannot edit internal tables with this form.") if $$fktable{'internal'};
   FKong::db::SendSQL("DELETE FROM ${tablesql} WHERE $pkey = ". FKong::db::SqlQuote($FKong::cgi::form{'id'}));
   FKong::db::SendSQL("DELETE FROM ${tablesql}_comment WHERE $pkey = ". FKong::db::SqlQuote($FKong::cgi::form{'id'})) if $$fktable{'comments'};
   FKong::db::SendSQL("DELETE FROM ${tablesql}_change WHERE $pkey = ". FKong::db::SqlQuote($FKong::cgi::form{'id'})) if $$fktable{'changes'};
   FKong::cgi::redirect(FKong::cgi::Uri(""));
}

#  The "list" is actually a string stored in a scalar
#  It neads to be cleaned for security and SQL compliance
sub clean_list 
{
   my($list) = @_;
   $list =~ tr/0-9,//dc;  # delete non-numeral non-commas  (redundant safety precaution)
   $list =~ s/,+/,/g;  # delete repeated commas
   $list =~ s/,$//g;  # delete trailing comma
   $list =~ s/^,//g;  # delete leading comma
   return $list;
}


$FKong::table_func{'list.html'} = \&show_list;
$FKong::table_func{'list.csv'} = $FKong::table_func{'list.html'};

sub show_list
{
   my($fktable) = @_;
   my $tablesql = $$fktable{'tablesql'};
   if($FKong::cgi::form{'delete_feature'}) {  
      delete_record($fktable);  
   } elsif(defined $FKong::cgi::form{'search'}) {  # if they are basic searching
      my $wheres = basic_query($fktable,'query');
      show_table("$$fktable{'tablesql'}.$$fktable{'pkey'},",$fktable,@$wheres);  
   } elsif($FKong::cgi::form{'text'}) {  # if they are super-basic searching (from the header)
      my $wheres = basic_query($fktable,'query');
      show_table("$$fktable{'tablesql'}.$$fktable{'pkey'},",$fktable,@$wheres);  
   } elsif(my $list = ($FKong::cgi::form{'list'} || FKong::session::get_state("$tablesql-list"))) {  # else if they have a saved list
      $list = clean_list($list);
      $FKong::cgi::keyword{'oldQuery'} = (FKong::session::get_state("$tablesql-query") || "");
      $FKong::cgi::keyword{'theList'} = $list;
      show_table("",$fktable,"$tablesql.$$fktable{'pkey'} IN ($list)");
   } else {  # else redirect them to the search form
      force_save_column_choice($fktable) if $KFong::cgi::form{'change_columns'};
      #show_table("$tablesql.$$fktable{'pkey'},$fktable,");  # as a default for now show them all
      FKong::cgi::redirect("search.html?");  #  last resort show search form
   }
}

$FKong::table_func{'count.html'} = \&show_counts;
$FKong::table_func{'count.csv'} = $FKong::table_func{'count.html'};

sub show_counts
{
   my($fktable) = @_;
   my $tablesql = $$fktable{'tablesql'};
   if($FKong::cgi::form{'count'}) {  # if they are basic searching
      push(@search_params,'vert','horiz','hrange','vrange');  
      my $wheres = basic_query($fktable,'count-query');
      FKong::count::show_count(1,$fktable,@$wheres);  
   } elsif(my $list = ($FKong::cgi::form{'list'} || FKong::session::get_state("$tablesql-list"))) {  # else if they have a saved list
      $list = clean_list($list);
      $FKong::cgi::keyword{'oldQuery'} = (FKong::session::get_state("$tablesql-count-query") || "");
      $FKong::cgi::keyword{'theList'} = $list;
      FKong::count::show_count(0,$fktable,"$tablesql.$$fktable{'pkey'} IN ($list)");
   } else {  # else redirect them to the search form
      FKong::cgi::redirect("count_form.html?");  #  last resort show count form
   }
}

$FKong::table_func{'columns.html'} = \&column_choice;

sub column_choice
{
   my($fktable) = @_;
   $FKong::cgi::keyword{'formurl'} = "list.html";
   my $tbl = FKong::list->new($$fktable{'tablesql'},$$fktable{'recprefix'},$$fktable{'url'}, {});
   define_list_object($tbl,$fktable);
   $FKong::cgi::keyword{'table'} = FKong::list::column_choice($tbl,$$fktable{'tablesql'});
   FKong::cgi::print_expanded_template("template/column_choice-template.html");
}

sub Done
{
   @search_params = ();
}

1;

__END__



